@@ -34,6 +34,10 @@
 *     [13] Cortes, David.
 *          "Isolation forests: looking beyond tree depth."
 *          arXiv preprint arXiv:2111.11639 (2021).
+*     [14] Ting, Kai Ming, Yue Zhu, and Zhi-Hua Zhou.
+*          "Isolation kernel and its effect on SVM"
+*          Proceedings of the 24th ACM SIGKDD
+*          International Conference on Knowledge Discovery & Data Mining. 2018.
 * 
 *     BSD 2-Clause License
 *     Copyright (c) 2019-2022, David Cortes
@@ -137,13 +141,56 @@ void merge_models(IsoForest*     model,      IsoForest*     other,
         throw std::runtime_error("Model to append trees to has indexer, but model to take trees from doesn't.\n");
     if (indexer != NULL && ind_other != NULL)
     {
-        if (!indexer->indices.front().node_distances.empty() && ind_other->indices.front().node_distances.empty())
+        bool indexer_is_empty = indexer->indices.empty();
+        bool ind_other_is_empty = ind_other->indices.empty();
+        bool model_is_empty = (model != NULL && model->trees.empty()) || (ext_model != NULL && ext_model->hplanes.empty());
+        bool other_is_empty = (other != NULL && other->trees.empty()) || (ext_model != NULL && ext_other->hplanes.empty());
+
+        if (indexer_is_empty && !model_is_empty && ind_other_is_empty && !other_is_empty) {
+            indexer = NULL;
+            ind_other = NULL;
+            goto skip_indexers;
+        }
+
+        if (!model_is_empty && !indexer_is_empty && !other_is_empty && ind_other_is_empty)
+            throw std::runtime_error("Model to append trees to has indexer, but model to take trees from doesn't.\n");
+        if (!model_is_empty && indexer_is_empty && !other_is_empty && !ind_other_is_empty)
+            throw std::runtime_error("Model to take trees from has indexer, but model to append trees to doesn't.\n");
+
+        if (
+            !indexer_is_empty && !ind_other_is_empty &&
+            indexer->indices.front().reference_points.size() != ind_other->indices.front().reference_points.size()
+        ) {
+            throw std::runtime_error("Model to append trees to and model to take trees from have different number of reference points.\n");
+        }
+
+
+        if (
+            !indexer_is_empty &&
+            !ind_other_is_empty &&
+            !indexer->indices.front().node_distances.empty() &&
+            ind_other->indices.front().node_distances.empty()
+        ) {
             throw std::runtime_error("Model to append trees to has indexer with distances, but model to take trees from has indexer without distances.\n");
-        if (!indexer->indices.front().reference_points.empty() && ind_other->indices.front().reference_points.empty())
+        }
+        if (
+            !indexer_is_empty &&
+            !ind_other_is_empty &&
+            !indexer->indices.front().reference_points.empty() &&
+            ind_other->indices.front().reference_points.empty()
+        ) {
             throw std::runtime_error("Model to append trees to has indexer with reference points, but model to take trees from has indexer without reference points.\n");
-        if (!indexer->indices.front().reference_indptr.empty() && ind_other->indices.front().reference_indptr.empty())
+        }
+        if (
+            !indexer_is_empty &&
+            !ind_other_is_empty &&
+            !indexer->indices.front().reference_indptr.empty() &&
+            ind_other->indices.front().reference_indptr.empty()
+        ) {
             throw std::runtime_error("Model to append trees to has indexer with kernel reference points, but model to take trees from has indexer without kernel reference points.\n");
+        }
     }
+    skip_indexers:
 
     try
     {
@@ -152,7 +199,7 @@ void merge_models(IsoForest*     model,      IsoForest*     other,
             if (model == other)
             {
                 auto other_copy = *other;
-                merge_models(model, &other_copy, ext_model, ext_other, imputer, iother, indexer, ind_other);
+                merge_models(model, &other_copy, NULL, NULL, NULL, NULL, NULL, NULL);
                 return;
             }
             model->trees.insert(model->trees.end(),
@@ -166,7 +213,7 @@ void merge_models(IsoForest*     model,      IsoForest*     other,
             if (ext_model == ext_other)
             {
                 auto other_copy = *ext_other;
-                merge_models(model, other, ext_model, &other_copy, imputer, iother, indexer, ind_other);
+                merge_models(NULL, NULL, ext_model, &other_copy, NULL, NULL, NULL, NULL);
                 return;
             }
             ext_model->hplanes.insert(ext_model->hplanes.end(),
@@ -179,7 +226,7 @@ void merge_models(IsoForest*     model,      IsoForest*     other,
             if (imputer == iother)
             {
                 auto other_copy = *iother;
-                merge_models(model, other, ext_model, ext_other, imputer, &other_copy, indexer, ind_other);
+                merge_models(NULL, NULL, NULL, NULL, imputer, &other_copy, NULL, NULL);
                 return;
             }
             imputer->imputer_tree.insert(imputer->imputer_tree.end(),
@@ -192,7 +239,7 @@ void merge_models(IsoForest*     model,      IsoForest*     other,
             if (indexer == ind_other)
             {
                 auto other_copy = *ind_other;
-                merge_models(model, other, ext_model, ext_other, imputer, iother, indexer, &other_copy);
+                merge_models(NULL, NULL, NULL, NULL, NULL, NULL, indexer, &other_copy);
                 return;
             }
             indexer->indices.insert(indexer->indices.end(),
@@ -201,7 +248,7 @@ void merge_models(IsoForest*     model,      IsoForest*     other,
         }
     }
 
-    catch(...)
+    catch (...)
     {
         if (model != NULL) model->trees.resize(curr_size_model);
         if (ext_model != NULL) ext_model->hplanes.resize(curr_size_model_ext);
